# CherryPy
# Copyright (C) 2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import importlib.resources

import cherrypy
from cherrypy.test import helper
from parameterized import parameterized

from .. import StaticMiddleware


class Static:

    components = StaticMiddleware()


class Root:

    static = Static()

    @cherrypy.expose
    def index(self):
        return "OK"


class StaticMiddlewareTest(helper.CPWebCase):
    default_lang = None
    interactive = False

    @classmethod
    def setup_server(cls):
        cherrypy.config.update(
            {
                'tools.i18n.default': cls.default_lang,
                'tools.i18n.mo_dir': importlib.resources.files(__package__) / 'locales',
                'tools.i18n.domain': 'messages',
            }
        )
        cherrypy.tree.mount(Root(), '/')

    @parameterized.expand(
        [
            '/static/components/Datatable.css',
            '/static/components/Datatable.js',
            '/static/components/Field.css',
            '/static/components/Field.js',
            '/static/components/SideBySideMultiSelect.css',
            '/static/components/SideBySideMultiSelect.js',
            '/static/components/Typeahead.css',
            '/static/components/Typeahead.js',
            '/static/components/vendor/bootstrap5/css/bootstrap.min.css',
            '/static/components/vendor/bootstrap5/js/bootstrap.min.js',
            '/static/components/vendor/datatables-extensions/Buttons/css/buttons.dataTables.min.css',
            '/static/components/vendor/datatables-extensions/Buttons/js/buttons.html5.min.js',
            '/static/components/vendor/datatables-extensions/Buttons/js/dataTables.buttons.min.js',
            '/static/components/vendor/datatables-extensions/FixedHeader/css/fixedHeader.dataTables.css',
            '/static/components/vendor/datatables-extensions/FixedHeader/js/dataTables.fixedHeader.min.js',
            '/static/components/vendor/datatables-extensions/JSZip/jszip.min.js',
            '/static/components/vendor/datatables-extensions/pdfmake/build/pdfmake.min.js',
            '/static/components/vendor/datatables-extensions/pdfmake/build/vfs_fonts.js',
            '/static/components/vendor/datatables-extensions/Responsive/css/responsive.dataTables.min.css',
            '/static/components/vendor/datatables-extensions/Responsive/js/dataTables.responsive.min.js',
            '/static/components/vendor/datatables-extensions/rowgroup/css/rowGroup.dataTables.min.css',
            '/static/components/vendor/datatables-extensions/rowgroup/js/dataTables.rowGroup.js',
            '/static/components/vendor/datatables/css/dataTables.dataTables.css',
            '/static/components/vendor/datatables/js/dataTables.min.js',
            '/static/components/vendor/jquery/jquery.min.js',
            '/static/components/vendor/popper/popper.min.js',
            '/static/components/vendor/typeahead/jquery.typeahead.min.css',
            '/static/components/vendor/typeahead/jquery.typeahead.min.js',
            '/static/components/vendor/typeahead/jquery.typeahead.min.js',
        ]
    )
    def test_static(self, url):
        self.getPage(url)
        self.assertStatus(200)



# SQLAlchemy plugins for cherrypy
# Copyright (C) 2022-2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import logging
import re

import cherrypy
from cherrypy.process.plugins import SimplePlugin
from sqlalchemy import create_engine, event, inspect
from sqlalchemy.engine import Engine
from sqlalchemy.exc import IntegrityError, SQLAlchemyError
from sqlalchemy.orm import declarative_base, scoped_session, sessionmaker


@event.listens_for(Engine, 'connect')
def _set_sqlite_journal_mode_wal(connection, connection_record):
    """
    Enable WAL journaling for concurrent read and write operation.
    """
    if 'sqlite3' in str(connection.__class__):
        cursor = connection.cursor()
        cursor.execute('PRAGMA journal_mode=WAL;')
        cursor.execute("PRAGMA foreign_keys=ON;")
        cursor.close()


@event.listens_for(Engine, "handle_error")
def enhance_database_errors(exception_context):
    """
    Enhance SQLAlchemy error by adding more information.
    """
    if isinstance(exception_context.sqlalchemy_exception, IntegrityError):
        constraint = _find_constraint(exception_context.sqlalchemy_exception)
        exception_context.sqlalchemy_exception.constraint = constraint


def _find_constraint(exception):
    """
    Search reference of constraint within the error message. Return None if not found.
    """
    # Extract the constraint name for Postgresql and SQLite
    # Postgresql: duplicate key value violates unique constraint "subnet_name_key"\nDETAIL:  Key (name)=() already exists.\n
    # Postgresql: violates check constraint "dnsrecord_value_domain_name"
    # Postgresql: foreign key constraint "dnsrecord_dnszone_subnet_fk"
    # SQLite: UNIQUE constrain: subnet.name
    # SQLite: UNIQUE constraint failed: index 'dnszone_name_index'
    # SQLite: CHECK constraint failed: dnsrecord_value_domain_name
    error = exception._message()
    constraint_match = (
        re.search(r'unique constraint "([^"]+)"', error)
        or re.search(r'check constraint "([^"]+)"', error)
        or re.search(r"UNIQUE constraint failed: index '([^']+)'", error)
        or re.search(r"UNIQUE constraint failed: (.+)$", error)
        or re.search(r"CHECK constraint failed: (.+)", error)
        or re.search(r'foreign key constraint "([^"]+)"', error)
    )
    if not constraint_match:
        return None
    name = constraint_match[1]

    # Use a lookup cache to simplify the search of index and constraints.
    if not getattr(_find_constraint, '_cache', False):
        cache = {}
        metadata = cherrypy.db.get_base().metadata
        for table in metadata.tables.values():
            for item in table.constraints:
                if item.name:
                    cache[item.name] = item
            for item in table.indexes:
                # Keep reference to unique index only.
                if item.unique:
                    if item.name:
                        cache[item.name] = item
                    # SQLite return <table>.<column>
                    key = ', '.join([f'{table.name}.{c.name}' for c in item.columns])
                    cache[key] = item
        _find_constraint._cache = cache

    return _find_constraint._cache.get(name, None)


def _get_model_changes(model):
    """
    Return a dictionary containing changes made to the model since it was
    fetched from the database.

    The dictionary is of the form {'property_name': [old_value, new_value]}
    """
    state = inspect(model)
    changes = {}
    for attr in state.attrs:
        hist = attr.history
        if not hist.has_changes():
            continue
        if isinstance(attr.value, (list, tuple)) or len(hist.deleted) > 1 or len(hist.added) > 1:
            # If array, store array
            changes[attr.key] = [hist.deleted, hist.added]
        else:
            # If primitive, store primitive
            changes[attr.key] = [
                hist.deleted[0] if len(hist.deleted) >= 1 else None,
                hist.added[0] if len(hist.added) >= 1 else None,
            ]
    return changes


class Base:
    '''
    Extends declarative base to provide convenience methods to models similar to
    functionality found in Elixir. Works in python3.

    For example, given the model User:
    # no need to write init methods for models, simply pass keyword arguments or
    # override if needed.
    User(name="daniel", email="daniel@dasa.cc").add()
    User.query # returns session.query(User)
    User.query.all() # instead of session.query(User).all()
    changed = User.from_dict({}) # update record based on dict argument passed in and returns any keys changed
    '''

    def add(self):
        """
        Add current object to session.
        """
        self.__class__.session.add(self)
        return self

    def delete(self):
        self.__class__.session.delete(self)
        return self

    def commit(self):
        self.__class__.session.commit()
        return self

    def flush(self):
        self.__class__.session.flush()
        return self

    def expire(self):
        self.__class__.session.expire(self)
        return self

    def rollback(self):
        self.__class__.session.rollback()
        return self


class SQLA(SimplePlugin):
    uri = None
    debug = False

    _base = None
    _session = None
    _engine = None

    @property
    def engine(self):
        return self._engine

    def start(self):
        if self.uri is None:
            return
        # Adjust debug level.
        if self.debug:
            logging.getLogger('sqlalchemy.engine').setLevel(logging.DEBUG)
        # Create connection to database
        self._engine = create_engine(self.uri)
        # Clean-up previous session.
        self.clear_sessions()
        # Associate our session to our engine
        self.get_session().configure(bind=self._engine)
        self.bus.log("database session plugin started")

    # This is slightly lower priority to get database started first.
    start.priority = 45

    def stop(self):
        if self._session:
            self.clear_sessions()
        if self._engine:
            self._engine.dispose()
        self.bus.log("database session plugin stopped")

    def graceful(self):
        """Reload of subscribers."""
        self.stop()
        self.start()

    def create_all(self):
        try:
            # Create tables
            base = self.get_base()
            conn = self.get_session().connection()
            base.metadata.create_all(bind=conn)
            self.get_session().commit()
        finally:
            # Release opened sessions.
            self.clear_sessions()

    def drop_all(self):
        try:
            # Drop all
            base = self.get_base()
            base.metadata.drop_all(bind=self._engine)
            self.get_session().commit()
        finally:
            # Release opened sessions.
            self.clear_sessions()

    def get_base(self):
        """
        Return a singleton instance of the Base classe for ORM.
        """
        if self._base is None:
            self._base = declarative_base(cls=Base)
            # Provide a friendly ObjectName.query.
            self._base.session = self.get_session()
            self._base.query = self.get_session().query_property()
        return self._base

    def get_session(self):
        """
        Return a singleton database session.
        """
        if self._session is None:
            self._session_factory = sessionmaker(autoflush=False, autocommit=False)
            self._session = scoped_session(self._session_factory)
        return self._session

    def after_request(self):
        self.clear_sessions()

    def clear_sessions(self):
        """
        Used to clean-up session and raise error if session are not clean.
        """
        if self._session is None:
            return
        try:
            # When terminating, raise an error if objects are not commit.
            if self._session.dirty or self._session.new or self._session.deleted:
                cherrypy.log(
                    'database session dirty; uncommitted objects detected — potential application bug',
                    context='DB',
                    severity=logging.ERROR,
                )
                if self._session.dirty:
                    changes = ', '.join([str(_get_model_changes(obj)) for obj in self._session.dirty])
                    cherrypy.log(
                        f'database session dirty_objects={self._session.dirty}', context='DB', severity=logging.ERROR
                    )
                    cherrypy.log(f'database session pending_changes={changes}', context='DB', severity=logging.ERROR)
                if self._session.new:
                    cherrypy.log(
                        f'database session new_objects={self._session.new}', context='DB', severity=logging.ERROR
                    )
                if self._session.deleted:
                    cherrypy.log(
                        f'database session deleted_objects={self._session.deleted}',
                        context='DB',
                        severity=logging.ERROR,
                    )
                raise SQLAlchemyError('session is dirty')
        finally:
            self._session.rollback()
            self._session.expunge_all()
            self._session.remove()


cherrypy.db = SQLA(cherrypy.engine)
cherrypy.db.subscribe()

cherrypy.config.namespaces['db'] = lambda key, value: setattr(cherrypy.db, key, value)



# Cherrypy-foundation
# Copyright (C) 2020-2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import cherrypy
from markupsafe import Markup
from wtforms.form import Form


class _ProxyFormdata:
    """
    Custom class to proxy default form data into WTForm from cherrypy variables.
    """

    def __contains__(self, key):
        return key in cherrypy.request.params

    def getlist(self, key):
        # Default to use cherrypy params.
        params = cherrypy.request.params
        if key in params:
            if isinstance(params[key], list):
                return params[key]
            else:
                return [params[key]]
        # Return default empty list.
        return []

    def __iter__(self):
        return iter(cherrypy.request.params)


_AUTO = _ProxyFormdata()


class CherryForm(Form):
    """
    Custom implementation of WTForm for cherrypy to support kwargs parms.

    If ``formdata`` is not specified, this will use cherrypy.request.params
    Explicitly pass ``formdata=None`` to prevent this.
    """

    def __init__(self, **kwargs):
        # Seamlessly support Json input if available.
        if 'json' in kwargs and kwargs.pop('json'):
            cherrypy.request.params = getattr(cherrypy.request, 'json', cherrypy.request.params)
        # Support explicit formdata
        if 'formdata' in kwargs:
            formdata = kwargs.pop('formdata')
        else:
            formdata = _AUTO if CherryForm.is_submitted(self) else None
        super().__init__(formdata=formdata, **kwargs)

    def is_submitted(self):
        """
        Consider the form submitted if there is an active request and
        the method is ``POST``.
        """
        return cherrypy.request.method == 'POST'

    def strict_validate(self):
        """
        Special validation to verify if all the field submited exists in this form.
        Raise an error if some fields are unknown.
        """
        form_errors = self.form_errors if hasattr(self, 'form_errors') else self.errors.setdefault(None, [])
        for key in cherrypy.request.params.keys():
            if key not in self:
                form_errors.append("unsuported field: %s" % key)
                return False
        return self.validate()

    def validate_on_submit(self):
        """
        Call `validate` only if the form is submitted.
        This is a shortcut for ``form.is_submitted() and form.validate()``.
        """
        return self.is_submitted() and self.validate()

    @property
    def error_message(self):
        """
        Return all error message in a single string.
        """
        if self.errors:
            msg = Markup("")
            for field, messages in self.errors.items():
                if msg:
                    msg += Markup('<br/>')
                # Field name
                if field in self:
                    msg += "%s: " % self[field].label.text
                elif field:
                    msg += "%s: " % field
                for m in messages:
                    msg += m
            return msg

    def populate_obj(self, obj):
        """
        Override default implementation to take acount of readonly fields.
        """
        for name, field in self._fields.items():
            if field.render_kw and field.render_kw.get('readonly'):
                continue
            field.populate_obj(obj, name)



# Cherrypy-foundation
# Copyright (C) 2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import importlib
from unittest import skipUnless
from urllib.parse import urlencode

import cherrypy
from cherrypy.test import helper
from parameterized import parameterized
from wtforms.fields import BooleanField, PasswordField, StringField, SubmitField
from wtforms.validators import InputRequired, Length

import cherrypy_foundation.tools.jinja2  # noqa
from cherrypy_foundation.form import CherryForm
from cherrypy_foundation.tools.i18n import gettext_lazy as _

HAS_JINJAX = importlib.util.find_spec("jinjax") is not None

env = cherrypy.tools.jinja2.create_env(
    package_name=__package__,
)


class LoginForm(CherryForm):
    login = StringField(
        _('User'),
        validators=[
            InputRequired(),
            Length(max=256, message=_('User too long.')),
        ],
        render_kw={
            "placeholder": _('User'),
            "autocorrect": "off",
            "autocapitalize": "none",
            "autocomplete": "off",
            "autofocus": "autofocus",
        },
    )
    password = PasswordField(
        _('Password'),
        validators=[
            InputRequired(),
            Length(max=256, message=_('Password too long.')),
        ],
        render_kw={"placeholder": _("Password")},
    )
    persistent = BooleanField(
        _('Remember me'),
        # All `label-*` are assigned to the label tag.
        render_kw={'container_class': 'col-sm-6', 'label-attr': 'FOO'},
    )
    submit = SubmitField(
        _('Login'),
        # All `container-*` are assigned to the container tag.
        render_kw={"class": "btn-primary float-end", 'container_class': 'col-sm-6', 'container-attr': 'BAR'},
    )


@cherrypy.tools.sessions(locking='explicit')
@cherrypy.tools.jinja2(env=env)
class Root:

    @cherrypy.expose
    def index(self, **kwargs):
        if 'login' not in cherrypy.session:
            raise cherrypy.HTTPRedirect('/login')
        return 'OK'

    @cherrypy.expose
    @cherrypy.tools.jinja2(template='test_form.html')
    def login(self, **kwargs):
        form = LoginForm()
        if form.validate_on_submit():
            # login user with cherrypy.tools.auth
            cherrypy.session['login'] = True
            raise cherrypy.HTTPRedirect('/')
        return {'form': form}


@skipUnless(HAS_JINJAX, reason='Required jinjax')
class FormTest(helper.CPWebCase):
    default_lang = None
    interactive = False

    @classmethod
    def setup_server(cls):
        cherrypy.tree.mount(Root(), '/')

    def test_get_form(self):
        # Given a form
        # When querying the page that include this form
        self.getPage("/login")
        self.assertStatus(200)
        # Then each field is render properly.
        # 1. Check title
        self.assertInBody('test-form')
        # 2. Check user field
        self.assertInBody('<label class="form-label" for="login">User</label>')
        self.assertInBody(
            '<input autocapitalize="none" autocomplete="off" autocorrect="off" autofocus="autofocus" class="form-control" id="login" maxlength="256" name="login" placeholder="User" required type="text" value="">'
        )
        # 3. Check password
        self.assertInBody('<label class="form-label" for="password">Password</label>')
        self.assertInBody(
            '<input class="form-control" id="password" maxlength="256" name="password" placeholder="Password" required type="password" value="">'
        )
        # 4 Check remember me
        self.assertInBody(
            '<input class="form-check-input" container-class="col-sm-6" id="persistent" label-attr="FOO" name="persistent" type="checkbox" value="y">'
        )
        self.assertInBody('<label attr="FOO" class="form-check-label" for="persistent">Remember me</label>')
        # 5. check submit button (regex matches because class could have different order with jinjax<=0.57)
        self.assertInBody('<div  attr="BAR"')
        self.assertMatchesBody(
            '<input class="(btn-primary ?|float-end ?|btn ?){3}" container-attr="BAR" container-class="col-sm-6" id="submit" name="submit" type="submit" value="Login">'
        )

    @parameterized.expand(
        [
            ('myuser', 'mypassword', 0, 303, False),
            ('myuser', '', 0, 200, 'Password: This field is required.'),
            ('', 'mypassword', 0, 200, 'User: This field is required.'),
        ]
    )
    def test_post_form(self, login, password, persistent, expect_status, expect_error):
        # Given a page with a form.
        # When data is sent to the form.
        self.getPage(
            "/login", method='POST', body=urlencode({'login': login, 'password': password, 'persistent': persistent})
        )
        # Then page return a status
        self.assertStatus(expect_status)
        # Then page may return an error
        if expect_error:
            self.assertInBody(expect_error)



# Scheduler plugins for Cherrypy
# Copyright (C) 2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import importlib
import tempfile
from threading import Event

import cherrypy
from cherrypy.test import helper

from .. import scheduler  # noqa

done = Event()

HAS_SQLALCHEMY = importlib.util.find_spec("sqlalchemy") is not None
if not HAS_SQLALCHEMY:
    pass
else:
    from sqlalchemy import Column, Integer, String
    from sqlalchemy.exc import IntegrityError

    from .. import db  # noqa

    Base = cherrypy.db.get_base()

    class User2(Base):
        __tablename__ = 'users2'
        id = Column(Integer, primary_key=True)
        username = Column(String)

        def __repr__(self):
            return f"User2(id={self.id}, username='{self.username}')"

    class Root:

        @cherrypy.expose
        def index(self):
            return str(User2.query.all())

        @cherrypy.expose
        def add(self, username):
            try:
                User2(username=username).add().commit()
                return "OK"
            except IntegrityError as e:
                return str(e)

    def create_user(*args, **kwargs):
        user = User2(*args, **kwargs).add()
        user.commit()
        done.set()


class DbSchedulerPluginTest(helper.CPWebCase):
    interactive = False

    @classmethod
    def setup_class(cls):
        cls.tempdir = tempfile.TemporaryDirectory(prefix='cherrypy-foundation-', suffix='-db-test')
        super().setup_class()

    @classmethod
    def teardown_class(cls):
        cls.tempdir.cleanup()
        super().teardown_class()

    @classmethod
    def setup_server(cls):
        cherrypy.config.update(
            {
                'db.uri': f"sqlite:///{cls.tempdir.name}/data.db",
            }
        )
        cherrypy.tree.mount(Root(), '/')

    def setUp(self) -> None:
        done.clear()
        cherrypy.db.create_all()
        return super().setUp()

    def tearDown(self):
        cherrypy.db.drop_all()
        return super().tearDown()

    def test_add_job_now(self):
        # Given a task
        # When scheduling that task
        scheduled = cherrypy.engine.publish('scheduler:add_job_now', create_user, username='myuser')
        self.assertTrue(scheduled)
        # When waiting for all tasks
        cherrypy.scheduler.wait_for_jobs()
        # Then the task get called
        self.assertTrue(done.is_set())
        # Then database was updated
        User2.query.filter(User2.username == 'myuser').one()



# SMTP Plugins for cherrypy
# Copyright (C) 2022-2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from unittest import mock, skipUnless

import cherrypy
from cherrypy.test import helper
from parameterized import parameterized

from .. import smtp  # noqa


class SmtpPluginTest(helper.CPWebCase):
    @classmethod
    def setup_server(cls):
        cherrypy.config.update(
            {
                'smtp.server': '__default__',
                'smtp.username': 'username',
                'smtp.password': 'password',
                'smtp.email_from': 'Test <email_from@test.com>',
            }
        )

    def test_send_mail(self):
        # Given a valid smtp server
        with mock.patch(smtp.__name__ + '.smtplib') as smtplib:
            # When publishing a send_mail
            cherrypy.engine.publish('send_mail', to='target@test.com', subject='subjet', message='body')
            # Then smtplib is called to send the mail.
            smtplib.SMTP.assert_called_once_with('__default__', 25)
            smtplib.SMTP.return_value.send_message.assert_called_once_with(mock.ANY)
            smtplib.SMTP.return_value.quit.assert_called_once_with()

    def test_send_mail_with_to_tuple(self):
        # Given a valid smtp server
        with mock.patch(smtp.__name__ + '.smtplib') as smtplib:
            # When publishing a send_mail
            cherrypy.engine.publish(
                'send_mail',
                to=('A name', 'target@test.com'),
                subject='subjet',
                message='body',
                bcc=('A bcc name', 'bcc@test.com'),
                reply_to=('A Reply Name', 'replyto@test.com'),
            )
            # Then smtplib is called to send the mail.
            smtplib.SMTP.assert_called_once_with('__default__', 25)
            smtplib.SMTP.return_value.send_message.assert_called_once_with(mock.ANY)
            smtplib.SMTP.return_value.quit.assert_called_once_with()

    @skipUnless(hasattr(cherrypy, 'scheduler'), reason='Required scheduler')
    def test_queue_mail(self):
        with mock.patch(smtp.__name__ + '.smtplib') as smtplib:
            # Given a mail being queued.
            cherrypy.engine.publish('queue_mail', to='target@test.com', subject='subjet', message='body')
            # When waiting for all task to be processed
            cherrypy.scheduler.wait_for_jobs()
            # Then smtplib is called to send the mail.
            smtplib.SMTP.assert_called_once_with('__default__', 25)
            smtplib.SMTP.return_value.send_message.assert_called_once_with(mock.ANY)
            smtplib.SMTP.return_value.quit.assert_called_once_with()

    def test_html2plaintext(self):
        """
        Check if this convertion is working fine.
        """

        html = """<html>
  <head>
    <style type="text/css">
      body { font-family:Helvetica; }
    </style>
  </head>
  <body>
    <h1>Hi!</h1>
    <p id="test"
        class="mb-2 text-center">
      How are you?<br/>
      Here&#160;is the <a href="https://www.python.org">link</a> you wanted.
    </p>
  </body>
</html>
"""

        expected = """**Hi!**
How are you?
Here is the link [1] you wanted.

[1] https://www.python.org"""
        self.assertEqual(expected, smtp._html2plaintext(html))

    def test_formataddr(self):
        self.assertEqual('test@test.com', smtp._formataddr('test@test.com'))
        self.assertEqual('TEST <test@test.com>', smtp._formataddr(('TEST', 'test@test.com')))
        self.assertEqual(
            'test2@test.com, TEST3 <test3@test.com>', smtp._formataddr(['test2@test.com', ('TEST3', 'test3@test.com')])
        )

    @parameterized.expand(
        [
            (None, None, None),
            (None, 'test2@test.com', 'test2@test.com'),
            ('test2@test.com', None, 'test2@test.com'),
            ('test1@test.com', 'test2@test.com', 'test1@test.com, test2@test.com'),
        ]
    )
    def test_with_bcc(self, smtp_bcc, bcc, expected_bcc):
        # Given a valid smtp server
        with mock.patch(smtp.__name__ + '.smtplib') as smtplib:
            # Given bcc is defined at plugin level.
            cherrypy.smtp.bcc = smtp_bcc
            # When publishing a send_mail with Bcc
            cherrypy.engine.publish(
                'send_mail',
                to=('A name', 'target@test.com'),
                subject='subjet',
                message='body',
                bcc=bcc,
            )
            # Then message is sent
            smtplib.SMTP.assert_called_once_with('__default__', 25)
            smtplib.SMTP.return_value.send_message.assert_called_once_with(mock.ANY)
            msg = smtplib.SMTP.return_value.send_message.call_args.args[0]
            smtplib.SMTP.return_value.quit.assert_called_once_with()
            # Message include our expected Bcc
            self.assertEqual(expected_bcc, msg['Bcc'])



# Cherrypy-foundation
# Copyright (C) 2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import cherrypy
from cherrypy.test import helper

import cherrypy_foundation.tools.jinja2  # noqa

from ..url import url_for

env = cherrypy.tools.jinja2.create_env(
    package_name=__package__,
    globals={
        'url_for': url_for,
    },
)


class SubPage:

    @cherrypy.expose
    @cherrypy.tools.jinja2(template='test_url.html')
    def index(self, **kwargs):
        _relative = cherrypy.request.headers.get('-Relative')
        return {'kwargs': {'_relative': _relative or None}}


@cherrypy.tools.proxy(base='https://www.example.com')
class ProxiedPage:

    @cherrypy.expose
    @cherrypy.tools.jinja2(template='test_url.html')
    def index(self, **kwargs):
        _relative = cherrypy.request.headers.get('-Relative')
        return {'kwargs': {'_relative': _relative or None}}


@cherrypy.tools.jinja2(env=env)
class Root:
    sub_page = SubPage()
    proxied = ProxiedPage()

    @cherrypy.expose
    @cherrypy.tools.jinja2(template='test_url.html')
    def index(self, **kwargs):
        _relative = cherrypy.request.headers.get('-Relative')
        return {'kwargs': {'_relative': _relative or None}}


class UrlTest(helper.CPWebCase):
    default_lang = None
    interactive = False

    @classmethod
    def setup_server(cls):
        cherrypy.tree.mount(Root(), '/')

    def test_url_for(self):
        self.assertEqual(url_for("foo", "bar"), 'http://127.0.0.1:54583/foo/bar')
        self.assertEqual(url_for("foo", "bar", _relative='server'), '/foo/bar')
        # Outside a request, relative url doesn't make alot of sens.
        self.assertEqual(url_for("foo", "bar", _relative=1), '127.0.0.1:54583/foo/bar')
        self.assertEqual(url_for("foo", "bar", _base='http://test.com'), 'http://test.com/foo/bar')

    def test_get_page(self):
        # Given a form
        # When querying the page that include this form
        self.getPage("/")
        self.assertStatus(200)
        # Then each field is render properly.
        # 1. Check title
        self.assertInBody('test-url')
        # 2. Check user field
        self.assertInBody(f'Empty: http://{self.HOST}:{self.PORT}/')
        self.assertInBody(f'Dot: http://{self.HOST}:{self.PORT}/')
        self.assertInBody(f'Dot page: http://{self.HOST}:{self.PORT}/my-page')
        self.assertInBody(f'Slash: http://{self.HOST}:{self.PORT}/')
        self.assertInBody(f'Page: http://{self.HOST}:{self.PORT}/my-page')
        self.assertInBody(f'Slash page: http://{self.HOST}:{self.PORT}/my-page')
        self.assertInBody(f'Query: http://{self.HOST}:{self.PORT}/my-page?bar=test+with+space&amp;foo=1')

    def test_get_page_relative_true(self):
        # Given a form
        # When querying the page that include this form
        self.getPage("/", headers=[('_relative', '1')])
        self.assertStatus(200)
        # Then each field is render properly.
        # 1. Check title
        self.assertInBody('test-url')
        # 2. Check user field
        self.assertInBody('Empty: <br/>')
        self.assertInBody('Dot: <br/>')
        self.assertInBody('Dot page: my-page<br/>')
        self.assertInBody('Slash: <br/>')
        self.assertInBody('Page: my-page<br/>')
        self.assertInBody('Slash page: my-page<br/>')
        self.assertInBody('Query: my-page?bar=test+with+space&amp;foo=1<br/>')

    def test_get_page_relative_server(self):
        # Given a form
        # When querying the page that include this form
        self.getPage("/", headers=[('_relative', 'server')])
        self.assertStatus(200)
        # Then each field is render properly.
        # 1. Check title
        self.assertInBody('test-url')
        # 2. Check user field
        self.assertInBody('Empty: /<br/>')
        self.assertInBody('Dot: /<br/>')
        self.assertInBody('Dot page: /my-page<br/>')
        self.assertInBody('Slash: /<br/>')
        self.assertInBody('Page: /my-page<br/>')
        self.assertInBody('Slash page: /my-page<br/>')
        self.assertInBody('Query: /my-page?bar=test+with+space&amp;foo=1<br/>')

    def test_get_page_proxied(self):
        # Given a form
        # When querying the page that include this form
        self.getPage("/proxied/")
        self.assertStatus(200)
        # Then each field is render properly.
        # 1. Check title
        self.assertInBody('test-url')
        # 2. Check user field
        self.assertInBody('Empty: https://www.example.com/')
        self.assertInBody('Dot: https://www.example.com/proxied/')
        self.assertInBody('Dot page: https://www.example.com/proxied/my-page<br/>')
        self.assertInBody('Slash: https://www.example.com/')
        self.assertInBody('Page: https://www.example.com/my-page')
        self.assertInBody('Slash page: https://www.example.com/my-page')
        self.assertInBody('Query: https://www.example.com/my-page?bar=test+with+space&amp;foo=1')

    def test_get_sub_page(self):
        # Given a form
        # When querying the page that include this form
        self.getPage("/sub-page/")
        self.assertStatus(200)
        # Then each field is render properly.
        # 1. Check title
        self.assertInBody('test-url')
        # 2. Check user field
        self.assertInBody(f'Empty: http://{self.HOST}:{self.PORT}/sub-page/')
        self.assertInBody(f'Dot: http://{self.HOST}:{self.PORT}/sub-page/')
        self.assertInBody(f'Dot page: http://{self.HOST}:{self.PORT}/sub-page/my-page')
        self.assertInBody(f'Slash: http://{self.HOST}:{self.PORT}/')
        self.assertInBody(f'Page: http://{self.HOST}:{self.PORT}/my-page')
        self.assertInBody(f'Slash page: http://{self.HOST}:{self.PORT}/my-page')
        self.assertInBody(f'Query: http://{self.HOST}:{self.PORT}/my-page?bar=test+with+space&amp;foo=1')



# Cherrypy-foundation
# Copyright (C) 2026 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''
Created on Apr. 10, 2020

@author: Patrik Dufresne
'''

import importlib
import unittest

from ..passwd import check_password, hash_password

HAS_ARGON2 = importlib.util.find_spec("argon2") is not None


class Test(unittest.TestCase):

    @unittest.skipUnless(HAS_ARGON2, "argon2 not installed")
    def test_check_password_argon(self):
        self.assertTrue(hash_password('admin12').startswith('$argon2'))
        self.assertTrue(
            check_password('admin123', '$argon2id$v=19$m=102400,t=2,p=8$/mDhOg8wyZeMTUjcbIC7mg$3pxRSfYgUXmKEKNtasP1Og')
        )

    @unittest.skipIf(HAS_ARGON2, "argon2 is installed")
    def test_check_password_hashlib(self):
        # Given argon is not installed
        # When generating hash password
        hash = hash_password('admin12')
        # Then SSHA hash is generated
        self.assertTrue(hash.startswith('{SSHA}'))

    def test_check_password(self):
        self.assertTrue(check_password('admin123', '{SSHA}/LAr7zGT/Rv/CEsbrEndyh27h+4fLb9h'))
        self.assertFalse(check_password('admin12', '{SSHA}/LAr7zGT/Rv/CEsbrEndyh27h+4fLb9h'))
        self.assertTrue(check_password('admin12', hash_password('admin12')))
        self.assertTrue(check_password('admin123', hash_password('admin123')))
